import pytest

from nanocurrency.blocks import Block

from tests.data import BLOCKS


@pytest.fixture(autouse=True)
def low_pow_threshold(monkeypatch):
    # Use a far lower default threshold for unit tests
    TEST_THRESHOLD = 9459044173002835106
    monkeypatch.setattr("nanocurrency.blocks.WORK_THRESHOLD", TEST_THRESHOLD)
    monkeypatch.setattr("nanocurrency.work.WORK_THRESHOLD", TEST_THRESHOLD)
    monkeypatch.setattr("nanocurrency.WORK_THRESHOLD", TEST_THRESHOLD)


@pytest.fixture(scope="function")
def block_factory():
    def _create_func(name):
        return Block.from_dict(BLOCKS[name]["data"])

    return _create_func



from decimal import Decimal

import pytest
from nanocurrency.units import convert, NanoDenomination


def test_floats_not_allowed():
    with pytest.raises(TypeError):
        convert(1.5, source=NanoDenomination.NANO, target=NanoDenomination.RAW)


@pytest.mark.parametrize("raw,knano", [
    (123456789, Decimal("0.000000000000000000123456789")),
    (1, Decimal("0.000000000000000000000000001")),
    (Decimal("123456789"), Decimal("0.000000000000000000123456789")),
    (Decimal("1"), Decimal("0.000000000000000000000000001")),
])
def test_knano_conversions(raw, knano):
    assert convert(
        raw, source=NanoDenomination.RAW, target=NanoDenomination.KILONANO
    ) == knano
    assert convert(
        knano, source=NanoDenomination.KILONANO, target=NanoDenomination.RAW
    ) == raw


@pytest.mark.parametrize("raw,mnano", [
    (123456789, Decimal("0.000000000000000000000123456789")),
    (1, Decimal("0.000000000000000000000000000001")),
    (Decimal("123456789"), Decimal("0.000000000000000000000123456789")),
    (Decimal("1"), Decimal("0.000000000000000000000000000001")),
])
def test_mnano_conversions(raw, mnano):
    assert convert(
        raw, source=NanoDenomination.RAW, target=NanoDenomination.MEGANANO
    ) == mnano
    assert convert(
        mnano, source=NanoDenomination.MEGANANO, target=NanoDenomination.RAW
    ) == raw


@pytest.mark.parametrize("raw,nano", [
    (123456789, Decimal("0.000000000000000123456789")),
    (1, Decimal("0.000000000000000000000001")),
    (Decimal("123456789"), Decimal("0.000000000000000123456789")),
    (Decimal("1"), Decimal("0.000000000000000000000001")),
])
def test_nano_conversions(raw, nano):
    assert convert(
        raw, source=NanoDenomination.RAW, target=NanoDenomination.NANO
    ) == nano
    assert convert(
        nano, source=NanoDenomination.NANO, target=NanoDenomination.RAW
    ) == raw


def test_max_nano_value():
    with pytest.raises(ValueError):
        # Value is larger than (2**128)-1 raw
        convert(
            Decimal("340282366.920938463463374607431768211456"),
            source=NanoDenomination.MEGANANO,
            target=NanoDenomination.RAW)

    with pytest.raises(ValueError):
        # Value is larger than (2**128)-1 raw
        convert(
            Decimal("340282366920938463463374607431768211456"),
            source=NanoDenomination.RAW,
            target=NanoDenomination.MEGANANO)

    convert(
        Decimal("340282366.920938463463374607431768211455"),
        source=NanoDenomination.MEGANANO,
        target=NanoDenomination.RAW
    ) == Decimal("340282366920938463463374607431768211455")

    convert(
        Decimal("340282366920938463463374607431768211455"),
        source=NanoDenomination.RAW,
        target=NanoDenomination.MEGANANO
    ) == Decimal("340282366.920938463463374607431768211455")



import hashlib
import random
import cpuinfo
import time

from binascii import hexlify, unhexlify
from concurrent.futures import ProcessPoolExecutor
from hashlib import blake2b

from bitarray import bitarray

from .exceptions import InvalidWork
from .util import is_hex

# Select the PoW C extension depending on highest supported instruction set
# based on the following priorities:
# AVX > SSE4.1 > SSSE3 > SSE2 > reference implementation
#
# This based on a Ryzen 1800X giving the following results:
# avx speed: 6185344 hashes/s (this is likely the fastest on Intel CPUs)
# sse4_1 speed: 6259267 hashes/s
# ssse3 speed: 6249287 hashes/s
# sse2 speed: 4635929 hashes/s
# ref speed: 4539306 hashes/s

# TODO: Maybe run a short benchmark when running solve_work() for the first
#       time?
_cpu_flags = cpuinfo.get_cpu_info()["flags"]

if "avx" in _cpu_flags:
    from . import _work_avx as _work
elif "sse4_1" in _cpu_flags:
    from . import _work_sse4_1 as _work
elif "ssse3" in _cpu_flags:  # SSSE3
    from . import _work_ssse3 as _work
elif "sse2" in _cpu_flags:  # SSE2
    from . import _work_sse2 as _work
else:
    from . import _work_ref as _work

WORK_THRESHOLD = int("ffffffc000000000", 16)


__all__ = (
    "WORK_THRESHOLD", "parse_work", "validate_work", "solve_work"
)


def parse_work(work):
    """
    Parses a proof-of-work (PoW) value and returns it if it's syntactically
    valid.

    .. note:: This method only checks that the work's format is correct.
              To validate a proof-of-work,
              use :func:`nanocurrency.work.validate_work`

    :param str work: Work as a 16-character hex string
    :raises InvalidWork: If the work is invalid
    :return: Work as a 16-character hex string
    :rtype: str
    """
    if not len(work) == 16 or not is_hex(work):
        raise InvalidWork("Work has to be a 16-character hexadecimal string")

    # NANO node represents PoW with lowercase hexadecimal letters for some
    # reason. Let's do the same here just to be sure.
    return work.lower()


def validate_work(block_hash, work, threshold=WORK_THRESHOLD):
    """Validate the proof-of-work.

    :param str block_hash: Block hash as a 64-character hex string
    :param str work: Work as a 16-character hex string
    :raises InvalidWork: If the work doesn't meet the threshold
    :return: The work as a 16-character hex string
    :rtype: str
    """
    reversed_work = bytearray(unhexlify(work))
    reversed_work.reverse()
    work_hash = bytearray(blake2b(
        b"".join([reversed_work, unhexlify(block_hash)]),
        digest_size=8).digest())
    work_hash.reverse()
    work_value = int(hexlify(work_hash), 16)

    if work_value < threshold:
        raise InvalidWork("Work doesn't meet the required threshold")
    else:
        return work.lower()


def solve_work(block_hash, threshold=WORK_THRESHOLD, timeout=None):
    """Solve the work for the corresponding block hash.

    :param str block_hash: Block hash as a 64-character hex string
    :param int threshold: The threshold/difficulty for the proof-of-work.
                          NANO network's threshold is used by default.
    :param timeout: Timeout in seconds. If provided, None will be returned
                    if the work can't be solved in the given time.
                    If None, the function will block until the work is solved.
    :type timeout: int, float or None
    :return: The solved work as a 64-character hex string or None
             couldn't be solved in time
    :rtype: str or None
    """
    # Reinitialize the random number generator in case this method is being
    # run in a multiprocessing environment; otherwise every process will
    # inherit the same RNG state
    random.seed()

    nonce = random.randint(0, (2**64)-1)
    block_hash_b = unhexlify(block_hash)

    start = time.time()

    while True:
        nonce = _work.do_work(block_hash_b, nonce, threshold)
        work = hexlify(int(nonce).to_bytes(8, byteorder="big"))
        try:
            validate_work(block_hash, work, threshold)
            return str(work, "utf-8")
        except InvalidWork:
            pass

        if timeout and (time.time() - start) > timeout:
            return None



import pytest

from nanocurrency.exceptions import (
    InvalidSeed, InvalidPrivateKey, InvalidPublicKey, InvalidAccount)
from nanocurrency.accounts import (
    AccountIDPrefix, get_account_key_pair, get_account_id,
    get_account_public_key, generate_account_private_key,
    generate_account_key_pair, generate_account_id, generate_seed,
    is_account_id_valid
)
from nanocurrency.util import is_hex


SEED = "bba817a4fa1418e10d014c99055c4922afa0f84b324e7850baf4b3b8b6af1a5b"
FIRST_PRIVATE_KEY = "64e0752579d3e6a6ea4f60b6596ff62cd59fc3160a687fe999522121d0a73c06"
FIRST_PUBLIC_KEY = "6ab5902b8e71b57f4b7680368eca010aa658affec3fd00adb4155dcee14efe29"
FIRST_ACCOUNT_ID_XRB = "xrb_1took1orwwfohx7qf13pju7144o8d4qzxizx14pua7cxsuinxzjbb6goarud"
FIRST_ACCOUNT_ID_NANO = "nano_1took1orwwfohx7qf13pju7144o8d4qzxizx14pua7cxsuinxzjbb6goarud"


def test_get_account_key_pair():
    with pytest.raises(InvalidPrivateKey):
        # Private key isn't 64 chars long
        get_account_key_pair("A"*63)

    with pytest.raises(InvalidPrivateKey):
        # Private key isn't in hex
        get_account_key_pair("G"*64)

    kp = get_account_key_pair(FIRST_PRIVATE_KEY)

    assert kp.private == FIRST_PRIVATE_KEY
    assert kp.public == FIRST_PUBLIC_KEY


def test_generate_account_private_key():
    with pytest.raises(InvalidSeed):
        # Seed isn't 64 chars long
        generate_account_private_key("A"*63, 0)

    with pytest.raises(InvalidSeed):
        # Seed isn't in hex
        generate_account_private_key("G"*64, 0)

    with pytest.raises(ValueError):
        # Index is not an integer
        generate_account_private_key(SEED, None)

    assert generate_account_private_key(SEED, 0) == FIRST_PRIVATE_KEY
    assert generate_account_private_key(SEED, 1) == \
        "69def30716080acafa63a257e513b02cffce6736a566fc89afcdc81b08e54d7d"
    assert generate_account_private_key(SEED, 10) == \
        "da45500aeb14c367fb9d8bb00c9e77152c418e1e6e472858856a72d9420859fd"


def test_generate_account_key_pair():
    with pytest.raises(InvalidSeed):
        # Seed isn't 64 chars long
        generate_account_key_pair("A"*63, 0)

    with pytest.raises(InvalidSeed):
        # Seed isn't in hex
        generate_account_key_pair("G"*64, 0)

    with pytest.raises(ValueError):
        # Index isn't an integer
        generate_account_key_pair(SEED, None)

    keypair = generate_account_key_pair(SEED, 0)

    assert keypair.private == FIRST_PRIVATE_KEY
    assert keypair.public == FIRST_PUBLIC_KEY


def test_get_account_id():
    """
    Test nanocurrency.accounts.get_account_id
    """
    # From PUBLIC KEY
    with pytest.raises(InvalidPublicKey):
        # Public key isn't 64 chars long
        get_account_id(public_key="A"*65)

    with pytest.raises(InvalidPublicKey):
        # Public key isn't in hex
        get_account_id(public_key="G"*64)

    assert get_account_id(public_key="0"*64) == \
        "xrb_1111111111111111111111111111111111111111111111111111hifc8npp"
    assert get_account_id(public_key=FIRST_PUBLIC_KEY) == FIRST_ACCOUNT_ID_XRB

    # From PRIVATE key
    with pytest.raises(InvalidPrivateKey):
        # Private key isn't 64 chars long
        get_account_id(private_key="A"*65)

    with pytest.raises(InvalidPrivateKey):
        # Private key isn't in hex
        get_account_id(private_key="G"*64)

    assert get_account_id(private_key="0"*64) == \
        "xrb_18gmu6engqhgtjnppqam181o5nfhj4sdtgyhy36dan3jr9spt84rzwmktafc"
    assert get_account_id(private_key=FIRST_PRIVATE_KEY) == \
        FIRST_ACCOUNT_ID_XRB

    # No/multiple parameters
    with pytest.raises(TypeError):
        # More than one parameter
        get_account_id(
            private_key=FIRST_PRIVATE_KEY, public_key=FIRST_PUBLIC_KEY)

    with pytest.raises(TypeError):
        # No parameters
        get_account_id(private_key=None, public_key=None)

    # Account ID prefix
    get_account_id(public_key=FIRST_PUBLIC_KEY) == FIRST_ACCOUNT_ID_XRB
    get_account_id(
        public_key=FIRST_PUBLIC_KEY, prefix=AccountIDPrefix.NANO
    ) == FIRST_ACCOUNT_ID_NANO

    with pytest.raises(ValueError):
        # Prefix is not 'nano_' or 'xrb_'
        get_account_id(public_key=FIRST_PUBLIC_KEY, prefix="fake_")


def test_get_account_public_key():
    """
    Test nanocurrency.accounts.get_account_public_key
    """
    # From ACCOUNT ID
    with pytest.raises(InvalidAccount):
        # Account ID doesn't start with 'xrb_1' or 'xrb_3'
        INVALID_ACCOUNT_ID = (
            "xrb_01111111111111111111111111111111111111111111111111111111111")
        get_account_public_key(account_id=INVALID_ACCOUNT_ID)

    with pytest.raises(InvalidAccount):
        # Account ID isn't 64 chars long
        get_account_public_key(account_id="a"*63)

    with pytest.raises(InvalidAccount):
        # Character '2' not a part of Base32
        INVALID_ACCOUNT_ID = \
            "xrb_122221orwwfohx7qf13pju7144o8d4qzxizx14pua7cxsuinxzjbb6goarud"
        get_account_public_key(account_id=INVALID_ACCOUNT_ID)

    with pytest.raises(InvalidAccount):
        # Account ID doesn't have a valid checksum
        INVALID_ACCOUNT_ID = \
            "xrb_1took1orwwfohx7qf13pju7144o8d4qzxizx14pua7cxsuinxzjbb6goarue"
        get_account_public_key(account_id=INVALID_ACCOUNT_ID)

    assert get_account_public_key(account_id=FIRST_ACCOUNT_ID_XRB) == \
        FIRST_PUBLIC_KEY

    # From PRIVATE KEY
    with pytest.raises(InvalidPrivateKey):
        # Private key isn't 64 chars long
        get_account_public_key(private_key="A"*63)

    with pytest.raises(InvalidPrivateKey):
        # Private key isn't in hex
        get_account_public_key(private_key="G"*64)

    assert get_account_public_key(private_key=FIRST_PRIVATE_KEY) == \
        FIRST_PUBLIC_KEY

    # No/multiple parameters
    with pytest.raises(TypeError):
        # More than one parameter
        get_account_public_key(
            private_key=FIRST_PRIVATE_KEY, account_id=FIRST_ACCOUNT_ID_XRB)

    with pytest.raises(TypeError):
        # No parameters
        get_account_public_key(private_key=None, account_id=None)


def test_generate_account_id():
    with pytest.raises(InvalidSeed):
        # Seed isn't 64 chars long
        generate_account_id("A"*63, 0)

    with pytest.raises(InvalidSeed):
        # Seed isn't in hex
        generate_account_id("G"*64, 0)

    with pytest.raises(ValueError):
        # Index isn't an integer
        generate_account_id(SEED, None)

    assert generate_account_id(SEED, 0) == FIRST_ACCOUNT_ID_XRB


def test_is_account_id_valid():
    assert not is_account_id_valid("invalid")
    assert not is_account_id_valid(
        "xrb_4took1orwwfohx7qf13pju7144o8d4qzxizx14pua7cxsuinxzjbb6goarud")
    assert not is_account_id_valid(
        "xrb_122221orwwfohx7qf13pju7144o8d4qzxizx14pua7cxsuinxzjbb6goarud")

    assert is_account_id_valid(FIRST_ACCOUNT_ID_XRB)


def test_generate_seed():
    seed = generate_seed()

    assert len(seed) == 64
    assert is_hex(seed)



from hashlib import blake2b

import pytest

import time

from nanocurrency.work import (
    parse_work, validate_work, solve_work
)
from nanocurrency.exceptions import InvalidWork


VALID_BLOCK_HASH = \
    "B585D9363B8265CFD5993F30A3D6DE6B5CA5CC7879E0AFA94D13F08B713B9FFD"
VALID_WORK = "5b064dcc70b9db0a"


def test_low_threshold_used_for_tests():
    """
    Check that a lower work threshold is used for tests
    """
    from nanocurrency.work import WORK_THRESHOLD

    assert WORK_THRESHOLD < int("ffffffc000000000", 16)


def test_parse_work():
    """
    Test nanocurrency.parse_work
    """
    with pytest.raises(InvalidWork):
        # Work has to be 16 chars long
        parse_work("A"*15)

    with pytest.raises(InvalidWork):
        # Woirk has to be in hex
        parse_work("x"*16)

    # parse_work returns lowercase hex string
    assert parse_work("A"*16) == "a"*16


def test_validate_work():
    with pytest.raises(InvalidWork):
        # Invalid work
        validate_work(block_hash=VALID_BLOCK_HASH, work="e"*16)

    assert validate_work(block_hash=VALID_BLOCK_HASH, work=VALID_WORK)


def test_solve_work():
    fake_hash = blake2b(b"fakeBlock", digest_size=32).hexdigest()

    # Use a very low threshold
    test_threshold = 1028

    result = solve_work(fake_hash, threshold=test_threshold)

    assert validate_work(fake_hash, work=result, threshold=test_threshold)


def test_work_timeout():
    """
    Try solving a work with timeout and make sure it times out
    after the time has passed
    """
    fake_hash = blake2b(b"fakeBlock", digest_size=32).hexdigest()

    start = time.time()

    # Use the maximum possible threshold to make the PoW essentially
    # impossible to solve
    result = solve_work(fake_hash, threshold=(2**64)-1, timeout=0.5)
    end = time.time()

    assert result is None

    # Operation should timeout after 0.5 seconds. Since it's a blocking
    # operation, give it some time to stop
    assert (end - start) > 0.5 and (end - start) < 1




import pytest

import json

from tests.data import BLOCKS

from nanocurrency.blocks import Block, MAX_BALANCE
from nanocurrency.exceptions import (
    InvalidAccount, InvalidBlock, InvalidSignature, InvalidWork,
    InvalidBlockHash, InvalidBalance, InvalidPublicKey)


VALID_ACCOUNT_ID = \
    "xrb_1took1orwwfohx7qf13pju7144o8d4qzxizx14pua7cxsuinxzjbb6goarud"
VALID_ACCOUNT_PUBLIC_KEY = \
    "6AB5902B8E71B57F4B7680368ECA010AA658AFFEC3FD00ADB4155DCEE14EFE29"

TEST_BLOCKS = []

for name in BLOCKS.keys():
    TEST_BLOCKS.append((name, BLOCKS[name]))


@pytest.mark.parametrize("name,block", TEST_BLOCKS)
def test_block_hash(name, block):
    """
    Calculate block hash for every test block and check that they match
    with the test data
    """
    correct_block_hash = block["hash"]
    block = Block.from_dict(block["data"])

    assert block.block_hash == correct_block_hash


@pytest.mark.parametrize("name,block", TEST_BLOCKS)
def test_block_json(name, block):
    """
    Deserialize every block from JSON and ensure they can be serialized
    back into identical JSON
    """
    test_block = block["data"]
    block = Block.from_json(json.dumps(test_block))

    assert json.loads(block.json()) == test_block


@pytest.mark.parametrize("name,block", TEST_BLOCKS)
def test_block_complete(name, block):
    """
    Deserialize every block from JSON and check that they have valid PoW and
    signatures
    """
    test_block = block["data"]
    block = Block.from_json(json.dumps(test_block))

    assert block.has_valid_work

    if block.account:
        assert block.has_valid_signature
        assert block.complete
    else:
        # Nano reference wallet's RPC API doesn't return some legacy blocks
        # with 'account' fields set even if they're required to verify the
        # signature.
        try:
            assert not block.account
            block.verify_signature()
        except ValueError as e:
            assert "'account' not included" in str(e)
            assert not block.complete


BLOCKS_AND_TYPES = (
    (BLOCKS["open"], "open"),
    (BLOCKS["send"], "send"),
    (BLOCKS["receive"], "receive"),
    (BLOCKS["change"], "change"),
    (BLOCKS["state_sendreceive"], "send/receive"),
    (BLOCKS["state_sendreceive2"], "send/receive"),
    (BLOCKS["state_change"], "change"),
)


def test_block_tx_type():
    """
    Load different blocks and check that their tx_types match
    """
    for block_data, tx_type in BLOCKS_AND_TYPES:
        block = Block.from_dict(block_data["data"])
        assert block.tx_type == tx_type, \
            "For block %s, expected tx_type %s, got %s" % (
                block_data["hash"], tx_type, block.tx_type)


def test_block_invalid_block_type():
    """
    Try creating a block with an invalid block type
    """
    with pytest.raises(ValueError) as exc:
        Block(block_type="transactionate_funds")

    assert "block_type" in str(exc)


def test_block_invalid_signature():
    """
    Try to load a block with invalid signature
    """
    # 'open' blocks are the only legacy blocks that can be verified as-is
    # Other blocks need to add the optional 'account' parameter
    block = BLOCKS["open"]["data"].copy()
    block["signature"] = "B"*128

    with pytest.raises(InvalidSignature) as exc:
        # Invalid signature (can't be verified)
        Block.from_dict(block)

    assert "couldn't be verified" in str(exc)

    block["signature"] = "A"

    with pytest.raises(InvalidSignature) as exc:
        # Invalid signature (incorrect length)
        Block.from_dict(block)

    assert "hexadecimal string" in str(exc)


def test_block_signature_with_missing_parameters():
    """
    Try to verify a signature for a block that's missing the 'account' or
    'signature' parameters

    NANO's rai_node API returns JSON representations of blocks that may not
    contain the 'account' parameter
    """
    block = Block.from_dict(BLOCKS["receive"]["data"])

    with pytest.raises(ValueError):
        # Missing account
        block.verify_signature()

    block.signature = None
    block.account = BLOCKS["receive"]["account"]

    with pytest.raises(ValueError):
        # Missing signature
        block.verify_signature()

    # Add the missing parameters
    block.account = BLOCKS["receive"]["account"]
    block.signature = BLOCKS["receive"]["data"]["signature"]

    block.verify_signature()


def test_block_sign():
    """
    Sign a block and verify it
    """
    private_key = \
        "587d4d70c1a3b66db89ad0e69e12bbd06774e8a161d2dca0c0c734556b8656ad"
    account_id = \
        "xrb_3f3iy4xh3umniqzpisrbuj6s1scde3mj83ffwgr4ckq1q6oirez1yjwq9t3y"

    block = Block(
        block_type="state",
        account=account_id,
        representative=account_id,
        previous=None,
        balance=100000,
        link="A4647CEBA216FD004AAFE3F552BA98739E6C4AD75A8C3E6A12B93531725D9F3A")

    assert not block.has_valid_signature

    block.sign(private_key)
    assert block.has_valid_signature

    with pytest.raises(ValueError):
        # Can't sign again if the block already has a signature
        block.sign(private_key)


def test_block_invalid_work(block_factory):
    """
    Try to load a block with invalid work
    """
    block_data = BLOCKS["receive"]["data"].copy()
    block_data["work"] = "a"*16

    with pytest.raises(InvalidWork):
        Block.from_dict(block_data)


def test_block_solve_work(block_factory):
    """
    Solve PoW for a work and verify it
    """
    block = block_factory("send")
    block.work = None

    assert not block.has_valid_work

    block.solve_work()
    assert block.has_valid_work

    with pytest.raises(ValueError):
        # Block already has a valid PoW
        block.solve_work()

    block.work = None

    # Try solving PoW with essentially impossible threshold
    assert not block.solve_work(threshold=(2**64)-1, timeout=0.1)


def test_block_verify_work(block_factory):
    """
    Load a block with work, and verify once with work and once without
    """
    block = block_factory("send")

    block.verify_work()

    with pytest.raises(InvalidWork):
        # Work doesn't meet the threshold
        block.work = "a"*16
        block.verify_work()

    with pytest.raises(ValueError):
        # Missing work
        block.work = None
        block.verify_work()


def test_block_skip_verify():
    """
    Deserialize a Block by skipping the verification of signature and PoW
    """
    block_data = BLOCKS["receive"]["data"].copy()
    block_data["work"] = "a"*16

    # The work is invalid, but it's OK
    Block.from_dict(block_data, verify=False)
    Block.from_json(json.dumps(block_data), verify=False)

    block_data["signature"] = "A"*128

    # The signature is invalid, but that's OK as well now
    Block.from_dict(block_data, verify=False)
    Block.from_json(json.dumps(block_data), verify=False)


def test_block_set_account(block_factory):
    """
    Test changing the 'account' block attribute
    """
    block = block_factory("receive")

    block.signature = None
    block.account = VALID_ACCOUNT_ID

    with pytest.raises(InvalidAccount):
        block.account = "invalid"


def test_block_set_source(block_factory):
    """
    Test changing the 'source' block attribute
    """
    block = block_factory("open")

    block.signature = None
    block.source = "A"*64

    with pytest.raises(InvalidBlockHash):
        block.source = "A"*65


def test_block_set_previous(block_factory):
    """
    Test changing the 'previous' block attribute
    """
    block = block_factory("send")

    block.signature = None
    block.previous = "A"*64

    with pytest.raises(InvalidPublicKey):
        block.previous = "A"*65


def test_block_set_destination(block_factory):
    """
    Test changing the 'destination' block attribute
    """
    block = block_factory("send")

    block.signature = None
    block.destination = VALID_ACCOUNT_ID

    with pytest.raises(InvalidAccount):
        block.destination = "invalid"


def test_block_set_representative(block_factory):
    """
    Test changing the 'representative' block attribute
    """
    block = block_factory("open")

    block.signature = None
    block.representative = VALID_ACCOUNT_ID

    with pytest.raises(InvalidAccount):
        block.destination = "invalid"


def test_block_set_balance(block_factory):
    """
    Test changing the 'balance' block attribute
    """
    # Hex-formatted balance
    block = block_factory("send")

    block.signature = None
    block.balance = 100

    assert json.loads(block.json())["balance"] == \
        "00000000000000000000000000000064"

    with pytest.raises(InvalidBalance):
        # Pre-state blocks only accept hex balances when deserializing
        # from a dict
        block_data = BLOCKS["send"]["data"].copy()
        block_data["balance"] = str(2**128)

        Block.from_dict(block_data)

    # String-formatted balance
    block = block_factory("state_sendreceive")

    block.signature = None
    block.balance = 100

    assert json.loads(block.json())["balance"] == "100"

    # Invalid balances
    with pytest.raises(InvalidBalance):
        block.balance = -1

    with pytest.raises(InvalidBalance):
        block.balance = MAX_BALANCE + 1


def test_block_legacy_send_from_dict():
    """
    When deserializing a legacy send block, the balance has to be hex-formatted
    """
    block_data = BLOCKS["send"]["data"].copy()
    block_data["balance"] = "10000000"

    with pytest.raises(InvalidBalance) as exc:
        Block.from_dict(block_data)

    assert "needs to be a hex" in str(exc)


def test_block_set_link(block_factory):
    """
    Test changing the 'link' block attribute which will also change
    'link_as_account'
    """
    block = block_factory("state_sendreceive")

    block.signature = None
    block.link = VALID_ACCOUNT_PUBLIC_KEY

    assert block.link == VALID_ACCOUNT_PUBLIC_KEY
    assert block.link_as_account == VALID_ACCOUNT_ID

    with pytest.raises(InvalidBlockHash):
        # Block hash in't hex
        block.link = "G"*64

    # None defaults to all-zeroes public key
    block.link = None

    assert block.link == "0" * 64
    assert block.link_as_account == \
        "xrb_1111111111111111111111111111111111111111111111111111hifc8npp"


def test_block_set_link_as_address(block_factory):
    """
    Test changing the 'link_as_account' block attribute which will also change
    'link'
    """
    block = block_factory("state_sendreceive")

    block.signature = None
    block.link_as_account = VALID_ACCOUNT_ID

    assert block.link_as_account == VALID_ACCOUNT_ID
    assert block.link == VALID_ACCOUNT_PUBLIC_KEY

    with pytest.raises(InvalidAccount):
        # 'link_as_account' isn't an account ID
        block.link_as_account = "invalid"

    # None defaults to all-zeroes public key
    block.link_as_account = None

    assert block.link == "0" * 64
    assert block.link_as_account == \
        "xrb_1111111111111111111111111111111111111111111111111111hifc8npp"


def test_block_missing_parameters():
    """
    Try to load a block with a missing required parameter
    """
    block_data = BLOCKS["receive"]["data"].copy()
    del block_data["previous"]

    with pytest.raises(InvalidBlock) as exc:
        Block.from_dict(block_data)

    assert "is missing required parameters: previous" in str(exc)


def test_block_prohibited_parameters():
    """
    Try to load a block with an additional prohibited parameter
    """
    block_data = BLOCKS["change"]["data"].copy()
    block_data["balance"] = "10000"

    with pytest.raises(InvalidBlock) as exc:
        Block.from_dict(block_data)

    assert "has prohibited parameters: balance" in str(exc)


def test_block_is_complete():
    """
    Check if a block is complete using Block.is_complete
    """
    block = Block.from_dict(BLOCKS["receive"]["data"])

    assert not block.complete

    # Add the 'account', so that the block can be verified
    block.account = BLOCKS["receive"]["account"]

    assert block.complete



import timeit
import cpuinfo

from hashlib import blake2b

from nanocurrency.work import _work


def run_speed_tests():
    def pow_solve_speed_test():
        ITERATIONS = 20

        # Get supported CPU instruction sets
        cpu_flags = cpuinfo.get_cpu_info()["flags"]

        all_flags = ["avx", "sse4_1", "ssse3", "sse2"]
        supported_flags = [
            flag for flag in all_flags
            if flag in cpu_flags
        ] + ["ref"]

        for flag in supported_flags:
            elapsed = timeit.Timer(
                ("nonce = 0\n"
                 "while nonce < 1000000:\n"
                 "    nonce = _work.do_work(block_hash, nonce, (2**64)-1)"),
                ("from nanocurrency import _work_{} as _work\n"
                 "from hashlib import blake2b\n"
                 "block_hash = blake2b(\n"
                 "    b'fakeBlock', digest_size=32).digest()\n".format(flag))
            ).timeit(ITERATIONS)

            # How many hashes per second?
            rate = int((1000000 * ITERATIONS) / elapsed)

            print(
                "BLAKE2b PoW ({flag}){used} speed: {rate} hashes/s".format(
                flag=flag,
                used=" DEFAULT" if supported_flags.index(flag) == 0 else "",
                rate=rate))

    def account_gen_speed_test():
        ITERATIONS = 10

        elapsed = timeit.Timer(
            ("i = 0\n"
             "while i < 1000:\n"
             "    a = generate_account_id(seed, i)\n"
             "    b = generate_account_key_pair(seed, i)\n"
             "    i += 1"),
            ("from nanocurrency.accounts import generate_seed, generate_account_id, generate_account_key_pair\n"
             "seed = generate_seed()")
        ).timeit(ITERATIONS)

        # How many account IDs per second?
        rate = int((1000 * ITERATIONS) / elapsed)

        print("Account generation rate: {} accounts/s".format(rate))

    def block_sign_speed_test():
        ITERATIONS = 2000

        elapsed = timeit.Timer(
            ("block.signature = None\n"
             "block.sign(key_pair.private)"),
            ("from nanocurrency.blocks import Block\n"
             "from nanocurrency.accounts import generate_seed, generate_account_id, generate_account_key_pair\n"
             "seed = generate_seed()\n"
             "account_id = generate_account_id(seed, 0)\n"
             "key_pair = generate_account_key_pair(seed, 0)\n"
             "block = Block(\n"
             "    block_type='open', account=account_id,\n"
             "    source='B2EC73C1F503F47E051AD72ECB512C63BA8E1A0ACC2CEE4EA9A22FE1CBDB693F',\n"
             "    representative='xrb_1anrzcuwe64rwxzcco8dkhpyxpi8kd7zsjc1oeimpc3ppca4mrjtwnqposrs')")
        ).timeit(ITERATIONS)

        rate = int(ITERATIONS / elapsed)

        print("Block sign rate: {} blocks/s".format(rate))

    def block_verify_speed_test():
        ITERATIONS = 2000

        elapsed = timeit.Timer(
            ("block.verify_signature()"),
            ("from nanocurrency.blocks import Block\n"
             "from nanocurrency.accounts import generate_seed, generate_account_id, generate_account_key_pair\n"
             "seed = generate_seed()\n"
             "account_id = generate_account_id(seed, 0)\n"
             "key_pair = generate_account_key_pair(seed, 0)\n"
             "block = Block(\n"
             "    block_type='open', account=account_id,\n"
             "    source='B2EC73C1F503F47E051AD72ECB512C63BA8E1A0ACC2CEE4EA9A22FE1CBDB693F',\n"
             "    representative='xrb_1anrzcuwe64rwxzcco8dkhpyxpi8kd7zsjc1oeimpc3ppca4mrjtwnqposrs')\n"
             "block.sign(key_pair.private)")
        ).timeit(ITERATIONS)

        rate = int(ITERATIONS / elapsed)

        print("Block verify rate: {} blocks/s".format(rate))

    pow_solve_speed_test()
    account_gen_speed_test()
    block_sign_speed_test()
    block_verify_speed_test()



from hashlib import blake2b

from ed25519_blake2b import VerifyingKey, SigningKey
from bitarray import bitarray

import binascii
import struct
import array
import random
import sys

__all__ = (
    "hex_to_uint4", "dec_to_hex", "uint_convert_precision",
    "uint4_to_bytes", "bytes_to_uint4", "bytes_to_uint5", "uint4_to_uint5",
    "uint5_to_uint4", "BASE32_LETTERS", "uint5_to_base32", "bytes_to_base32",
    "base32_to_uint5", "hex_to_base32", "is_hex"
)


def hex_to_uint4(h):
    h = "".join(["0" + c for c in h])
    return array.array("B", bytes.fromhex(h))


def dec_to_hex(d, n):
    return format(d, "0{}X".format(n*2))
    # return "%0.{bytes}X".format(bytes=n*2) % d


def uint_convert_precision(uints, source=4, dest=8):
    bits = bitarray()

    for v in uints:
        bits.extend("{{0:0{}b}}".format(source).format(v))

    dest_values = []

    bit_count = len(bits)
    for i in range(0, bit_count, dest):
        if (i+dest) > bit_count:
            continue

        dest_values.append(int(bits[i:i+dest].to01(), 2))

    return dest_values


def uint4_to_bytes(uint4):
    return bytes(uint_convert_precision(uint4, 4, 8))


def bytes_to_uint4(b):
    return uint_convert_precision(b, 8, 4)


def bytes_to_uint5(b):
    return uint_convert_precision(b, 8, 5)


def uint4_to_uint5(uint4):
    return uint_convert_precision(uint4, 4, 5)


def uint5_to_uint4(uint5):
    return uint_convert_precision(uint5, 5, 4)


BASE32_LETTERS = "13456789abcdefghijkmnopqrstuwxyz"


def uint5_to_base32(uint5):
    return "".join([BASE32_LETTERS[u] for u in uint5])


def bytes_to_base32(b):
    uint5 = bytes_to_uint5(b)

    return uint5_to_base32(uint5)


def base32_to_uint5(base32):
    return [BASE32_LETTERS.index(c) for c in base32]


def hex_to_base32(h):
    b = binascii.unhexlify(h)
    return bytes_to_base32(b)


def is_hex(h):
    try:
        int(h, 16)
        return True
    except ValueError:
        return False



import json
import os


BLOCKS_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "data", "blocks.json")


with open(BLOCKS_PATH, "r") as f:
    BLOCKS = json.loads(f.read())["blocks"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import glob
import io
import os
import sys
from shutil import rmtree

from setuptools import Command, Extension, setup
from setuptools.command.test import test as TestCommand

import versioneer

# Package meta-data.
NAME = 'pynanocurrency'
DESCRIPTION = \
    'Python library for working with the NANO cryptocurrency protocol'
URL = 'https://github.com/Matoking/pynanocurrency'
EMAIL = 'jannepulk@gmail.com'
AUTHOR = 'Janne Pulkkinen'
REQUIRES_PYTHON = '>=3.6.0'

# What packages are required for this module to be executed?
REQUIRED = [
    'bitarray>=0.8.1', 'ed25519-blake2b>=1.4', 'py-cpuinfo>=4'
]

NANOCURRENCY_WORK_REF = Extension(
    "nanocurrency._work_ref",
    include_dirs=[
        "src/nanocurrency-work-module/ref",
    ],
    sources=[
        "src/nanocurrency-work-module/work.c",
    ] + glob.glob("src/nanocurrency-work-module/ref/*.c"),
    extra_compile_args=["-DWORK_REF"]
)

NANOCURRENCY_WORK_SSE2 = Extension(
    "nanocurrency._work_sse2",
    include_dirs=[
        "src/nanocurrency-work-module/sse",
    ],
    sources=[
        "src/nanocurrency-work-module/work.c",
    ] + glob.glob("src/nanocurrency-work-module/sse/*.c"),
    extra_compile_args=["-DWORK_SSE2", "-msse2"]
)

NANOCURRENCY_WORK_SSSE3 = Extension(
    "nanocurrency._work_ssse3",
    include_dirs=[
        "src/nanocurrency-work-module/sse",
    ],
    sources=[
        "src/nanocurrency-work-module/work.c",
    ] + glob.glob("src/nanocurrency-work-module/sse/*.c"),
    extra_compile_args=["-DWORK_SSSE3", "-mssse3"]
)

NANOCURRENCY_WORK_SSE4_1 = Extension(
    "nanocurrency._work_sse4_1",
    include_dirs=[
        "src/nanocurrency-work-module/sse",
    ],
    sources=[
        "src/nanocurrency-work-module/work.c",
    ] + glob.glob("src/nanocurrency-work-module/sse/*.c"),
    extra_compile_args=["-DWORK_SSE4_1", "-msse4.1"]
)

NANOCURRENCY_WORK_AVX = Extension(
    "nanocurrency._work_avx",
    include_dirs=[
        "src/nanocurrency-work-module/sse",
    ],
    sources=[
        "src/nanocurrency-work-module/work.c",
    ] + glob.glob("src/nanocurrency-work-module/sse/*.c"),
    extra_compile_args=["-DWORK_AVX", "-mavx"]
)

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import shlex
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


class SpeedTest(Command):
    description = (
        "Run a set of speed tests to measure performance of cryptographic "
        "operations"
    )
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from tests.performance_test import run_speed_tests

        run_speed_tests()


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(versioneer.get_version()))
        os.system('git push --tags')

        sys.exit()


cmdclass = versioneer.get_cmdclass()
cmdclass.update({
    'upload': UploadCommand,
    'speed': SpeedTest,
    'pytest': PyTest,
})

# Where the magic happens:
setup(
    name=NAME,
    version=versioneer.get_version(),
    description=DESCRIPTION,
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    ext_modules=[
        NANOCURRENCY_WORK_REF,
        NANOCURRENCY_WORK_SSE2,
        NANOCURRENCY_WORK_SSSE3,
        NANOCURRENCY_WORK_SSE4_1,
        NANOCURRENCY_WORK_AVX,
    ],
    packages=["nanocurrency"],
    package_data={"": ["LICENSE"]},
    package_dir={"nanocurrency": "src/nanocurrency"},
    install_requires=REQUIRED,
    setup_requires=["sphinx"],
    tests_require=["pytest"],
    include_package_data=True,
    license='CC0',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        #  'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Operating System :: POSIX :: Linux',
        'Topic :: Office/Business :: Financial',
    ],
    # $ setup.py publish support.
    cmdclass=cmdclass,
    command_options={
        'build_sphinx': {
            'project': ('setup.py', NAME),
            'version': ('setup.py', versioneer.get_version()),
            'source_dir': ('setup.py', 'docs')
        },
    }
)
